__all__ = ["calc_epl", "get_spectra", "setup_receiver"]


# standard library
from collections import deque
from collections.abc import Iterator, Sequence
from contextlib import contextmanager
from datetime import datetime, timedelta
from logging import getLogger
from scipy.optimize import curve_fit
from struct import Struct
from threading import Event, Lock, Thread
from typing import Callable, NamedTuple


# dependencies
import numpy as np
import xarray as xr
from numpy.typing import NDArray
from ..vdif.receive import get_socket


# constants
C = 299_792_458  # m/s
LITTLE_ENDIAN: str = "<"
LOCK = Lock()
LOGGER = getLogger(__name__)
LOWER_FREQ_MHZ = 16384
N_BYTES_PER_SCAN: int = 1312 * 64
N_BYTES_PER_UNIT: int = 1312
N_CHANS_FOR_FORMAT = 2048
N_ROWS_CORR_DATA: int = 512
N_ROWS_CORR_HEAD: int = 64
N_ROWS_VDIF_HEAD: int = 8
N_UNITS_PER_SCAN: int = 64
PACKET_BUFFER = deque(maxlen=6000)
REF_EPOCH_ORIGIN = np.datetime64("2000", "Y")
REF_EPOCH_UNIT = np.timedelta64(6, "M")
SHORT: str = "h"
TIME_PER_SCAN: float = 1e-2
UINT: str = "I"


# global variables
FEED = ["c", "t", "r", "b", "l"]
UDP_READY_EVENT = Event()


@contextmanager
def setup_receiver(
    *,
    group: str = "239.0.0.1",
    port: int = 11111,
) -> Iterator[None]:
    """Setup the UDP receiver for EPL calculation.

    Args:
        group: Multicast group address.
        port: Multicast port number.

    """

    sock = get_socket(group=group, port=port)

    receiver_thread = Thread(
        target=udp_receiver,
        args=(sock, UDP_READY_EVENT),
        daemon=True,
    )

    try:
        LOGGER.debug("Starting receiver thread...")
        receiver_thread.start()
        yield
    finally:
        LOGGER.debug("Finishing receiver thread...")
        receiver_thread.join(timeout=1.0)
        sock.close()


def get_spectra(
    feed_origin: str,
    feed_pattern: Sequence[str],
    freq_range: tuple[float, float],
    freq_binning: int = 8,
    size: int = 25,
) -> xr.DataArray:
    """Get Spectra.

    Args:
        feed_pattern: Feed name pattern to be repeated.
        freq_range: Frequency range to select (in Hz).
        freq_binning: Number of frequency channels to bin.
        size: Number of scan.

    """

    feed_origin = datetime.strptime(feed_origin, "%Y%m%dT%H%M%S")  # type: ignore
    COUNT = np.zeros(5, dtype=int)
    ch = int((freq_range[1] - freq_range[0]) * 1e-6 / freq_binning)  # MHz
    SPECTRA = np.zeros((5, ch), dtype=np.complex128)
    FEED_PATTERN = feed_pattern
    FREQ = get_freq(freq_binning)  # Hz
    FREQ_SELECTED = FREQ[(FREQ >= freq_range[0]) & (FREQ <= freq_range[1])]

    UDP_READY_EVENT.clear()
    UDP_READY_EVENT.wait()
    UDP_READY_EVENT.clear()
    scan = get_latest_packets(size)

    for i in range(size):
        frames = scan[i]
        data_time, spectrum = get_nth_spectrum_in_range(
            frames, FREQ, freq_range, freq_binning
        )
        if i == 0:
            start_time = data_time

        n = get_n_from_current_time(feed_origin, data_time)  # type: ignore
        target = FEED_PATTERN[n % len(FEED_PATTERN)]
        if target == "x":
            continue

        f = FEED.index(target)
        SPECTRA[f] += spectrum
        COUNT[f] += 1

    last_time = data_time

    LOGGER.debug(
        f"start_time={start_time}, last_time={last_time}, d_time={last_time-start_time}"
    )

    return xr.DataArray(
        data=[
            np.array(SPECTRA[0] / COUNT[0]),
            np.array(SPECTRA[1] / COUNT[1]),
            np.array(SPECTRA[2] / COUNT[2]),
            np.array(SPECTRA[3] / COUNT[3]),
            np.array(SPECTRA[4] / COUNT[4]),
        ],
        dims=("feed", "freq"),
        coords={
            "feed": FEED,
            "freq": FREQ_SELECTED,
            "time": last_time,
        },
    )


def calc_epl(spec: xr.DataArray) -> xr.DataArray:
    freq = spec.coords["freq"].values

    epl_dict = {}
    for f in ["c", "t", "r", "b", "l"]:
        epl_dict[f] = get_epl(spec.sel(feed=f).data, freq)

    return xr.DataArray(
        data=[
            epl_dict["c"],
            epl_dict["t"],
            epl_dict["r"],
            epl_dict["b"],
            epl_dict["l"],
        ],
        dims="feed",
        coords={"feed": FEED, "time": spec.coords["time"]},
    )


def udp_receiver(sock, udp_ready_event):
    while True:
        temp_buffer = []
        while True:
            frame, _ = sock.recvfrom(N_BYTES_PER_UNIT)
            array = np.frombuffer(
                frame,
                dtype=[
                    ("word_0", "u4"),
                    ("word_1", "u4"),
                    ("word_2", "u4"),
                    ("word_3", "u4"),
                    ("word_4", "u4"),
                    ("word_5", "u4"),
                    ("word_6", "u4"),
                    ("word_7", "u4"),
                    ("data", ("u1", 1280)),
                ],
            )
            word_4 = Word(array["word_4"])
            ch = word_4[16:24]
            if ch == 64:
                break

        while True:
            frame, _ = sock.recvfrom(N_BYTES_PER_UNIT)

            if len(frame) != N_BYTES_PER_UNIT:
                LOGGER.warning(
                    f" Received frame size anomaly: {len(frame)} bytes. Sample skipped."
                )
                break
            temp_buffer.append(frame)

            if len(temp_buffer) == N_UNITS_PER_SCAN:
                if not check_channel_order(temp_buffer):
                    LOGGER.warning(
                        "Channel sequence mismatch detected. Reinitializing reception."
                    )
                    break
                with LOCK:
                    PACKET_BUFFER.append(list(temp_buffer))
                temp_buffer.clear()
                udp_ready_event.set()


class Word:
    """VDIF header word parser."""

    def __init__(self, data: NDArray[np.int_]):
        self.data = data

    """VDIF header word as a 1D integer array."""

    def __getitem__(self, index: slice, /) -> NDArray[np.int_]:
        """Slice the VDIF header word."""
        start, stop = index.start, index.stop
        return (self.data >> start) & ((1 << stop - start) - 1)


class head_data(NamedTuple):
    time: datetime
    thread_id: NDArray[np.int_]
    ch: NDArray[np.int_]
    integ: NDArray[np.int_]


def read_head(frame: bytes) -> head_data:
    array = np.frombuffer(
        frame,
        dtype=[
            ("word_0", "u4"),
            ("word_1", "u4"),
            ("word_2", "u4"),
            ("word_3", "u4"),
            ("word_4", "u4"),
            ("word_5", "u4"),
            ("word_6", "u4"),
            ("word_7", "u4"),
            ("data", ("u1", 1280)),
        ],
    )
    word_0 = Word(array["word_0"])
    word_1 = Word(array["word_1"])
    word_3 = Word(array["word_3"])
    word_4 = Word(array["word_4"])
    seconds = int(word_0[0:30])
    frame_num = int(word_1[0:24])
    ref_epoch = int(word_1[24:30])
    thread_id = int(word_3[16:26])
    ch = int(word_4[16:24])
    integ = int(word_4[0:8])

    time = (
        REF_EPOCH_ORIGIN
        + REF_EPOCH_UNIT * ref_epoch
        + np.timedelta64(1, "s") * seconds
        + np.timedelta64(integ * (frame_num // 64), "ms")
    )
    time_dt = time.astype("datetime64[us]").astype(datetime)

    return head_data(time_dt, thread_id, ch, integ)  # type: ignore


def get_latest_packets(a: int) -> list:
    with LOCK:
        return list(PACKET_BUFFER)[-a:]


def check_channel_order(packet_set: list[bytes]) -> bool:
    ch_list = []

    for frame in packet_set:
        metadata = read_head(frame)
        ch_list.append(int(metadata.ch))
    expected = list(range(1, 65))
    if ch_list == expected:
        return True
    else:
        return False


# main features
def get_spectrum(
    scan: xr.Dataset,
    freq_binning: int = 8,
) -> tuple[datetime, np.ndarray]:
    n_integ = 1
    n_units = N_UNITS_PER_SCAN * n_integ
    n_chans = N_ROWS_CORR_DATA // 2

    spectra = np.empty([n_units, n_chans], dtype=complex)

    for i in range(n_units):
        frame = scan[i]
        time = read_head(frame).time  # type: ignore
        corr_data = read_corr_data(frame[288:1312])
        spectra[i] = parse_corr_data(corr_data)

    spectra = spectra.reshape([n_integ, N_UNITS_PER_SCAN * n_chans])
    spectrum = integrate_spectra(spectra, freq_binning)
    return time, spectrum


def get_nth_spectrum_in_range(
    scan: xr.Dataset,
    freq: np.ndarray,
    freq_range: tuple[float, float],
    freq_binning: int = 8,
) -> tuple[datetime, np.ndarray]:
    time, spec = get_spectrum(scan, freq_binning)
    filtered_spec = spec[(freq >= freq_range[0]) & (freq <= freq_range[1])]
    return time, filtered_spec


def get_epl(spec: np.ndarray, freq: np.ndarray) -> float:
    fit = curve_fit(line_through_origin, freq, get_phase(spec))
    slope = fit[0]
    slope = slope[0]
    epl = (C * slope) / (2 * np.pi)
    return epl


def get_freq(bin_width: int = 8, n_chans: int = 2048) -> np.ndarray:
    freq = 1e6 * (LOWER_FREQ_MHZ + np.arange(n_chans * bin_width))
    freq = freq.reshape((freq.shape[0] // bin_width, bin_width)).mean(-1)
    return freq


def get_n_from_current_time(start_time: datetime, data_time: datetime) -> int:
    dt = (data_time - start_time).total_seconds()
    n = int(round(dt / TIME_PER_SCAN)) - 1
    return n


def get_amp(da: np.ndarray) -> np.ndarray:
    """複素数DataArrayの絶対値Amplitudeを返す関数"""
    amp = np.abs(da)
    return amp


def get_phase(da: np.ndarray) -> np.ndarray:
    """複素数DataArrayの偏角(ラジアン単位)を返す関数"""
    phase = np.arctan2(da.imag, da.real)
    return phase


def line_through_origin(freq: np.ndarray, slope: float) -> np.ndarray:
    """原点を通る直線モデル"""
    return slope * freq


def integrate_spectra(spectra: np.ndarray, freq_binning: int = 8) -> np.ndarray:
    spectrum = spectra.mean(0)
    return spectrum.reshape([len(spectrum) // freq_binning, freq_binning]).mean(1)


# struct readers
def make_binary_reader(n_rows: int, dtype: str) -> Callable:
    struct = Struct(LITTLE_ENDIAN + dtype * n_rows)

    def reader(f):
        if isinstance(f, bytes):
            return struct.unpack(f)
        else:
            return struct.unpack(f.read(struct.size))

    return reader


read_vdif_head: Callable = make_binary_reader(N_ROWS_VDIF_HEAD, UINT)
read_corr_head: Callable = make_binary_reader(N_ROWS_CORR_HEAD, UINT)
read_corr_data: Callable = make_binary_reader(N_ROWS_CORR_DATA, SHORT)


# struct parsers
def parse_vdif_head(vdif_head: list):
    # not implemented yet
    pass


def parse_corr_head(corr_head: list):
    # not implemented yet
    pass


# 相関データ
def parse_corr_data(corr_data: list) -> np.ndarray:
    real = np.array(corr_data[0::2])
    imag = np.array(corr_data[1::2])
    return real + imag * 1j



__all__ = ["Frames", "get_frames", "get_socket", "receive"]


# standard library
from collections import deque
from collections.abc import Iterator
from contextlib import contextmanager
from logging import getLogger
from os import PathLike
from socket import (
    AF_INET,
    INADDR_ANY,
    IPPROTO_IP,
    IP_ADD_MEMBERSHIP,
    SOCK_DGRAM,
    SOL_SOCKET,
    SO_REUSEADDR,
    inet_aton,
    socket,
)
from struct import pack
from threading import Event, Lock, Thread

try:
    from socket import SO_REUSEPORT  # type: ignore
except ImportError:
    SO_REUSEPORT = None  # type: ignore


# dependencies
from tqdm import tqdm
from . import FRAME_BYTES


# constants
LOGGER = getLogger(__name__)


class Frames:
    """Thread-safe deque for VDIF frames.

    Args:
        size: Maximum number of the latest frames to keep.

    """

    def __init__(self, size: int = 100000, /) -> None:
        self.data = deque(maxlen=size)
        self.lock = Lock()

    def append(self, frame: bytes, /) -> None:
        """Append a frame to the right end of the deque."""
        with self.lock:
            self.data.append(frame)

    def appendleft(self, frame: bytes, /) -> None:
        """Append a frame to the left end of the deque."""
        with self.lock:
            self.data.appendleft(frame)

    def get(self, n: int, /) -> list[bytes]:
        """Get the n frames from the right end of the deque."""
        with self.lock:
            n = min(n, len(self.data))
            return [self.data[i] for i in range(-n, 0)]

    def getleft(self, n: int, /) -> list[bytes]:
        """Get the n frames from the left end of the deque."""
        with self.lock:
            n = min(n, len(self.data))
            return [self.data[i] for i in range(0, n)]

    def pop(self, n: int, /) -> list[bytes]:
        """Pop the n frames from the right end of the deque."""
        with self.lock:
            n = min(n, len(self.data))
            return [self.data.pop() for _ in range(n)][::-1]

    def popleft(self, n: int, /) -> list[bytes]:
        """Pop the n frames from the left end of the deque."""
        with self.lock:
            n = min(n, len(self.data))
            return [self.data.popleft() for _ in range(n)]

    def __len__(self) -> int:
        """Get the number of frames in the deque."""
        with self.lock:
            return len(self.data)


def receive(
    vdif: PathLike[str] | str,
    /,
    *,
    group: str = "239.0.0.1",
    port: int = 11111,
    size: int = 100000,
    status: bool | int = False,
    usage: bool | int = False,
) -> None:
    """Receive a VDIF file over UDP multicast.

    Args:
        vdif: Path to the VDIF file.
        group: Multicast group address.
        port: Multicast port number.
        size: Maximum number of the latest frames to keep.
        status: Whether to show the receiving status.
        usage: Whether to show the buffer usage.

    """
    with get_frames(size, group=group, port=port, usage=usage) as frames:
        try:
            with (
                open(vdif, "wb") as file,
                tqdm(
                    desc=f"Receiving {vdif}",
                    disable=not status,
                    position=int(status) - 1,
                    unit="B",
                    unit_scale=True,
                ) as bar,
            ):
                while True:
                    try:
                        bar.update(file.write(frames.popleft(1)[0]))
                    except Exception:
                        pass
        except KeyboardInterrupt:
            LOGGER.warning("Receiving interrupted by user.")


@contextmanager
def get_frames(
    size: int = 100000,
    /,
    *,
    group: str = "239.0.0.1",
    port: int = 11111,
    usage: bool | int = False,
) -> Iterator[Frames]:
    """Get streamed VDIF frames over UDP multicast.

    Args:
        size: Maximum number of the latest frames to keep.
        group: Multicast group address.
        port: Multicast port number.
        usage: Whether to show the deque usage.

    Yields:
        Frames: Thread-safe deque for VDIF frames.

    """
    with get_socket(group=group, port=port) as sock:
        frames = Frames(size)
        event = Event()

        def receive():
            with tqdm(
                bar_format="{l_bar}{bar}| {n_fmt}/{total_fmt}",
                desc=f"Deque usage",
                disable=not usage,
                position=int(usage) - 1,
                total=size,
                unit="%",
            ) as bar:
                while not event.is_set():
                    frame, _ = sock.recvfrom(FRAME_BYTES)
                    frames.append(frame)
                    bar.n = len(frames)
                    bar.refresh()

        thread = Thread(target=receive, daemon=True)

        try:
            thread.start()
            yield frames
        finally:
            event.set()
            thread.join(1)


def get_socket(*, group: str = "239.0.0.1", port: int = 11111) -> socket:
    """Get a socket object for UDP multicast."""
    sock = socket(AF_INET, SOCK_DGRAM)
    sock.setsockopt(SOL_SOCKET, SO_REUSEADDR, 1)

    if SO_REUSEPORT is not None:
        sock.setsockopt(SOL_SOCKET, SO_REUSEPORT, 1)

    sock.bind(("", port))

    mreq = pack("4sL", inet_aton(group), INADDR_ANY)
    sock.setsockopt(IPPROTO_IP, IP_ADD_MEMBERSHIP, mreq)
    return sock



__all__ = [
    "Converter",
    "get_aggregated",
    "get_converter",
    "get_epl",
    "get_feed",
    "get_freq",
    "get_phase",
]


# standard library
from collections.abc import Sequence
from dataclasses import dataclass


# dependencies
import numpy as np
import xarray as xr
from ndtools import Range
from numpy.typing import NDArray
from decode.stats import mean
from ..utils import to_datetime, to_timedelta


# constants
C = 299_792_458  # m s^-1
FREQ_AT_CHAN0 = 16384e6  # Hz
FREQ_AT_CHAN8192 = 24576e6  # Hz
FREQ_STEP = 1e6  # Hz


@dataclass
class Converter:
    """Aggregated-to-EPL data converter for the Nobeyama 45m telescope.

    Args:
        cal_interval: Bandpass calibration interval (float in seconds
            or string with units). If None, calibration is done
            only once at the beginning.
        last: Last aggregated data used for the bandpass calibration.
            It should be None when a converter is created.

    """

    cal_interval: np.timedelta64 | str | float | None = None
    last: xr.DataArray | None = None

    def __call__(
        self, aggregated: xr.DataArray, /
    ) -> tuple[xr.DataArray, xr.DataArray]:
        """Get the EPL data (in m; feed) from the aggregated data.

        Args:
            aggregated: Aggregated data (feed x freq).

        Returns:
            Estimated EPL (in m; feed)
            Estimated EPL at calibration (in m; feed).

        """
        if self.last is None or (
            self.cal_interval is not None
            and aggregated.time - self.last.time > to_timedelta(self.cal_interval)
        ):
            self.last = aggregated
            self.count = 0

        self.count += 1
        return (
            get_epl(aggregated / self.last.data),
            get_epl(self.last / self.last.data),
        )


def get_aggregated(
    samples: xr.DataArray,
    /,
    *,
    elevation: float = 0.0,
    feed_origin: np.datetime64 | str | None = None,
    feed_pattern: Sequence[str] = ("",),
    freq_binning: int = 8,
    freq_range: Range = Range(19.5e9, 22.5e9),
) -> xr.DataArray:
    """Get aggregated data (feed x freq) from VDIF samples.

    Args:
        samples: VDIF samples (time x chan).
        elevation: Antenna elevation angle (in deg).
        feed_origin: Origin time for the feed name pattern.
        feed_pattern: Feed name pattern to be repeated.
        freq_binning: Number of frequency channels to bin.
        freq_range: Frequency range (in Hz) to select.

    Returns:
        Aggregated data (feed x freq).

    """
    aggregated = (
        samples.groupby(get_feed(samples, feed_pattern, feed_origin))
        .mean("time")
        .rename("aggregated")
        .assign_coords(
            elevation=xr.DataArray(elevation, attrs={"units": "deg"}),
            freq=get_freq(samples),
            time=samples.time[-1],
        )
        .swap_dims(chan="freq")
    )
    aggregated = mean(aggregated, dim={"freq": freq_binning})
    return aggregated.sel(freq=aggregated.freq == freq_range)


def get_converter(
    cal_interval: np.timedelta64 | str | float | None = None,
    /,
) -> Converter:
    """Get an aggregated-to-EPL data converter for the Nobeyama 45m telescope.

    Args:
        cal_interval: Bandpass calibration interval (float in seconds
            or string with units). If None, calibration is done
            only once at the beginning.

    Returns:
        Aggregated-to-EPL data converter.

    """
    return Converter(cal_interval=cal_interval)


def get_epl(aggregated: xr.DataArray, /) -> xr.DataArray:
    """Get the EPL data (in m; feed) from the aggregated data."""
    return (
        np.arctan2(aggregated.imag, aggregated.real)
        .curvefit("freq", get_phase)
        .get("curvefit_coefficients")
        .sel(param="epl")
        .drop_vars("param")
        .rename("epl")
        .assign_attrs(units="m")
    )


def get_feed(
    samples: xr.DataArray,
    pattern: Sequence[str] = ("",),
    origin: np.datetime64 | str | None = None,
    /,
) -> xr.DataArray:
    """Get the feed names from the VDIF samples.

    Args:
        samples: VDIF samples (time x chan).
        pattern: Feed name pattern to be repeated.
        origin: Origin time for the feed name pattern.

    Returns:
        Feed names of the VDIF samples (feed).

    """
    if origin is None:
        origin = samples.time[0]
    else:
        origin = to_datetime(origin)

    dt = np.timedelta64(samples.ip_length, "ms")
    index = ((samples.time - origin) / dt).astype(int)

    return xr.DataArray(
        np.array(pattern)[index % len(pattern)],
        dims="time",
        coords={"time": samples.time},
        name="feed",
    )


def get_freq(samples: xr.DataArray, /) -> xr.DataArray:
    """Get the frequency (in Hz) of the VDIF samples."""
    freq = FREQ_AT_CHAN0 + FREQ_STEP * samples.chan
    freq[freq >= FREQ_AT_CHAN8192] = np.nan
    return freq.rename("freq").assign_attrs(units="Hz")


def get_phase(freq: NDArray, epl: float, /) -> NDArray:
    """Get the phase (in rad) from frequency (in Hz) and EPL (in m)"""
    return 2 * np.pi * epl * freq / C



__all__ = ["send"]


# standard library
from logging import getLogger
from os import PathLike
from pathlib import Path
from socket import AF_INET, IPPROTO_IP, IP_MULTICAST_TTL, SOCK_DGRAM, socket
from time import perf_counter


# dependencies
from tqdm import trange
from . import FRAMES_PER_SAMPLE, FRAME_BYTES
from .convert import get_integ_time
from ..utils import sleep


# constants
LOGGER = getLogger(__name__)


def send(
    vdif: PathLike[str] | str,
    /,
    *,
    group: str = "239.0.0.1",
    port: int = 11111,
    repeat: bool = False,
    status: bool | int = False,
    ttl: int = 1,
) -> None:
    """Send a VDIF file over UDP multicast.

    Args:
        vdif: Path to the VDIF file.
        group: Multicast group address.
        port: Multicast port number.
        repeat: Whether to repeat sending the VDIF file.
        status: Whether to show the sending status.
        ttl: Time-to-live for multicast packets.

    """
    n_frames, remainder = divmod(Path(vdif).stat().st_size, FRAME_BYTES)

    if remainder:
        LOGGER.warning(f"VDIF file is truncated ({remainder} bytes remaining).")

    with socket(AF_INET, SOCK_DGRAM) as sock:
        sock.setsockopt(IPPROTO_IP, IP_MULTICAST_TTL, ttl)

        def send_once() -> None:
            with open(vdif, "rb") as file:
                for _ in trange(
                    n_frames,
                    desc=f"Sending {vdif}",
                    disable=not status,
                    position=int(status) - 1,
                    unit="frame",
                    unit_scale=True,
                ):
                    start = perf_counter()
                    sock.sendto(frame := file.read(FRAME_BYTES), (group, port))
                    seconds_per_frame = get_integ_time(frame) / FRAMES_PER_SAMPLE
                    end = perf_counter()
                    sleep(seconds_per_frame - (end - start))

        try:
            if repeat:
                while True:
                    send_once()
            else:
                send_once()
        except KeyboardInterrupt:
            LOGGER.warning("Sending interrupted by user.")



__all__ = [
    "FRAME_BYTES",
    "VDIF_HEAD_BYTES",
    "CORR_HEAD_BYTES",
    "CORR_DATA_BYTES",
    "CHANS_PER_FRAME",
    "FRAMES_PER_SAMPLE",
    "convert",
    "receive",
    "send",
]


# constants
FRAME_BYTES = 1312
VDIF_HEAD_BYTES = 32
CORR_HEAD_BYTES = 256
CORR_DATA_BYTES = 1024
CHANS_PER_FRAME = 256
FRAMES_PER_SAMPLE = 64


# dependencies
from . import convert
from . import receive
from . import send



__all__ = ["control"]


# standard library
from collections import deque
from collections.abc import Sequence
from logging import getLogger
from os import PathLike
from pathlib import Path
from time import sleep
from warnings import catch_warnings, simplefilter


# dependencies
import numpy as np
import xarray as xr
from ndtools import Range
from tqdm import tqdm
from .convert import get_converter as get_subref_converter
from ..cosmos import get_cosmos
from ..epl.convert import get_aggregated, get_converter as get_epl_converter
from ..vdif import FRAMES_PER_SAMPLE
from ..vdif.convert import get_samples
from ..vdif.receive import get_frames
from ..utils import log, take, to_timedelta


# constants
LOGGER = getLogger(__name__)
SECOND = np.timedelta64(1, "s")
TIME_UNITS = "microseconds since 2000-01-01T00:00:00"


def control(
    *,
    # options for the feed information
    feed_model: PathLike[str] | str,
    feed_origin: np.datetime64 | str,
    feed_pattern: Sequence[str] | str,
    # options for the EPL estimates
    cal_interval: np.timedelta64 | str | float = "10 s",
    freq_binning: int = 8,
    freq_range: tuple[float, float] = (19.5e9, 22.5e9),  # Hz
    integ_per_sample: np.timedelta64 | str | float = "0.01 s",
    integ_per_epl: np.timedelta64 | str | float = "0.5 s",
    # options for the subref control
    dry_run: bool = True,
    control_duration: np.timedelta64 | str | float = "1 d",
    control_period: np.timedelta64 | str | float = "0.5 s",
    epl_interval_tolerance: float = 0.1,
    integral_gain_dX: float = 0.1,
    integral_gain_dZ: float = 0.1,
    proportional_gain_dX: float = 0.1,
    proportional_gain_dZ: float = 0.1,
    range_ddX: tuple[float, float] = (0.00005, 0.000375),  # m
    range_ddZ: tuple[float, float] = (0.00005, 0.000300),  # m
    # options for network connection
    cosmos_host: str = "127.0.0.1",
    cosmos_port: int = 11111,
    cosmos_safe: bool = True,
    vdif_group: str = "239.0.0.1",
    vdif_port: int = 22222,
    # option for display
    status: bool = True,
    # options for data saving
    epl_data: PathLike[str] | str | None = None,
    epl_data_max: int | None = None,
    subref_data: PathLike[str] | str | None = None,
    subref_data_max: int | None = None,
    # options for logging
    log_file: PathLike[str] | str | None = None,
    log_file_level: int | str = "INFO",
    log_stderr: bool = True,
    log_stderr_level: int | str = "WARNING",
) -> None:
    """Control the subreflector of the Nobeyama 45m telescope by MAO."""
    with log(
        file=log_file,
        file_level=log_file_level,
        stderr=log_stderr,
        stderr_level=log_stderr_level,
    ):
        for key, val in locals().items():
            LOGGER.debug(f"{key}: {val!r}")

        # define the frame size for each EPL estimate
        dt_control = to_timedelta(control_duration)
        dt_epl = to_timedelta(integ_per_epl)
        dt_sample = to_timedelta(integ_per_sample)
        epl_size = int(dt_control / dt_epl)
        frame_size = FRAMES_PER_SAMPLE * int(dt_epl / dt_sample)

        # create the EPL and subref converters
        get_epl = get_epl_converter(cal_interval)
        get_subref = get_subref_converter(
            control_period=control_period,
            epl_interval_tolerance=epl_interval_tolerance,
            feed_model=feed_model,
            proportional_gain_dX=proportional_gain_dX,
            proportional_gain_dZ=proportional_gain_dZ,
            integral_gain_dX=integral_gain_dX,
            integral_gain_dZ=integral_gain_dZ,
            range_ddX=range_ddX,
            range_ddZ=range_ddZ,
        )

        with (
            tqdm(disable=not status, unit="EPL") as bar,
            get_cosmos(host=cosmos_host, port=cosmos_port, safe=cosmos_safe) as cosmos,
            get_frames(frame_size * 2, group=vdif_group, port=vdif_port) as frames,
        ):
            # wait until enough frames are buffered
            while len(frames.get(frame_size)) != frame_size:
                sleep(dt_epl / SECOND)

            epls = deque(maxlen=epl_data_max)
            subrefs = deque(maxlen=subref_data_max)

            try:
                for _ in range(epl_size):
                    with take(dt_epl / SECOND):
                        # get the current telescope state
                        state = cosmos.receive_state()

                        # get the current VDIF samples (time x chan)
                        frames_ = frames.get(frame_size + FRAMES_PER_SAMPLE)
                        samples = get_samples(frames_)

                        # get the aggregated data (feed x freq)
                        aggregated = get_aggregated(
                            samples,
                            elevation=state.elevation,
                            feed_pattern=feed_pattern,
                            feed_origin=feed_origin,
                            freq_binning=freq_binning,
                            freq_range=Range(*freq_range),
                        )

                        # estimate the EPL (in m; feed)
                        epl, epl_cal = get_epl(aggregated)

                        # estimate the current subref parameters
                        subref = get_subref(epl, epl_cal)

                        # send the subref control to COSMOS
                        if not dry_run:
                            cosmos.send_subref(
                                dX=float(subref.sel(drive="X")),
                                dZ=float(subref.sel(drive="Z")),
                            )

                        # append EPL and/or subref data (optional)
                        if epl_data is not None:
                            epls.append(epl)

                        if subref_data is not None:
                            subrefs.append(subref)

                        # update the progress bar
                        bar.update(1)
            except KeyboardInterrupt:
                LOGGER.warning("Control interrupted by user.")
            finally:
                # save EPL data (optional)
                if epl_data is not None:
                    encoding = {"time": {"units": TIME_UNITS}}
                    epls = xr.concat(epls, dim="time")

                    with catch_warnings():
                        simplefilter("ignore", category=UserWarning)

                        if Path(epl_data).exists():
                            epls.to_zarr(epl_data, mode="a", append_dim="time")
                        else:
                            epls.to_zarr(epl_data, mode="w", encoding=encoding)

                # save subref data (optional)
                if subref_data is not None:
                    encoding = {"time": {"units": TIME_UNITS}}
                    subrefs = xr.concat(subrefs, dim="time")

                    with catch_warnings():
                        simplefilter("ignore", category=UserWarning)

                        if Path(subref_data).exists():
                            subrefs.to_zarr(subref_data, mode="a", append_dim="time")
                        else:
                            subrefs.to_zarr(subref_data, mode="w", encoding=encoding)



__all__ = ["log", "sleep", "take", "to_datetime", "to_timedelta"]


# standard library
from collections.abc import Iterator
from contextlib import contextmanager
from datetime import datetime, timezone
from logging import FileHandler, Formatter, LogRecord, StreamHandler, getLogger
from os import PathLike
from time import perf_counter, sleep as sleep_


# dependencies
import numpy as np
import pandas as pd
from dateparser import parse


# constants
LOGGER = getLogger(__name__)
LOGGER_FORMAT = "{asctime} | {name} | {levelname} | {message}"


class ISO8601Formatter(Formatter):
    """Formatter whose default time format is ISO 8601."""

    def formatTime(self, record: LogRecord, datefmt: str | None = None) -> str:
        return datetime.fromtimestamp(record.created, timezone.utc).isoformat()


@contextmanager
def log(
    *,
    file: PathLike[str] | str | None = None,
    file_level: int | str = "DEBUG",
    stderr: bool = True,
    stderr_level: int | str = "INFO",
) -> Iterator[None]:
    """Context manager for the root logger configuration.

    Args:
        file: Path to the log file.
        file_level: Logging level for the log file.
        stderr: Whether to log to stderr.
        stderr_level: Logging level for stderr.

    """
    root = getLogger()
    current_level = root.level
    current_handlers = root.handlers.copy()
    formatter = ISO8601Formatter(LOGGER_FORMAT, style="{")

    try:
        root.setLevel("DEBUG")

        for handler in root.handlers:
            root.removeHandler(handler)

        if file is not None:
            handler = FileHandler(file)
            handler.setFormatter(formatter)
            handler.setLevel(file_level)
            root.addHandler(handler)

        if stderr:
            handler = StreamHandler()
            handler.setFormatter(formatter)
            handler.setLevel(stderr_level)
            root.addHandler(handler)

        yield
    finally:
        root.setLevel(current_level)

        for handler in root.handlers:
            root.removeHandler(handler)

        for handler in current_handlers:
            root.addHandler(handler)


def sleep(seconds: float, /) -> None:
    """Busy-waiting sleep (more precise but more CPU usage)."""
    start = perf_counter()
    end = start + seconds

    while perf_counter() < end:
        pass


@contextmanager
def take(duration: float, /, *, precise: bool = False) -> Iterator[None]:
    """Run a code block for a specified duration.

    Args:
        duration: Run time of the code block in seconds.
        precise: Whether to use busy-waiting sleep for more precise timing.

    """
    start = perf_counter()
    yield
    end = perf_counter()

    if (elapsed := end - start) > duration:
        LOGGER.warning(f"Block run exceeded {duration} s.")
    else:
        LOGGER.debug(f"Block run finished in {elapsed} s.")

        if precise:
            sleep(duration - elapsed)
        else:
            sleep_(duration - elapsed)


def to_datetime(value: np.datetime64 | str, /) -> np.datetime64:
    """Parse a string into a NumPy datetime64[ns] object in UTC."""
    if isinstance(value, np.datetime64):
        return value.astype("M8[ns]")

    if (parsed := parse(value)) is None:
        raise ValueError(f"Could not parse to datetime: {value!s}")

    return pd.to_datetime(parsed).tz_convert("UTC").to_datetime64()


def to_timedelta(value: np.timedelta64 | str | float, /) -> np.timedelta64:
    """Parse a string or float into a NumPy timedelta64[ns] object."""
    if isinstance(value, np.timedelta64):
        return value.astype("m8[ns]")

    if isinstance(value, str):
        return pd.to_timedelta(value).to_timedelta64()

    return pd.to_timedelta(value, "s").to_timedelta64()



__all__ = ["cosmos", "subref", "vdif", "utils"]
__version__ = "0.4.0"


# dependencies
from fire import Fire
from . import cosmos
from . import subref
from . import vdif
from . import utils


def main():
    Fire(
        {
            "cosmos": {
                "receive": cosmos.receive,
                "send": cosmos.send,
            },
            "subref": {
                "control": subref.control.control,
            },
            "vdif": {
                "receive": vdif.receive.receive,
                "send": vdif.send.send,
            },
        }
    )



__all__ = ["Cosmos", "State", "Subref", "get_cosmos", "get_socket", "receive", "send"]


# standard library
import re
from dataclasses import dataclass
from datetime import datetime
from logging import getLogger
from socket import AF_INET, SOCK_STREAM, socket
from zoneinfo import ZoneInfo


# dependencies
from typing_extensions import Self


# constants
ABSMAX_DX = 0.048  # m
ABSMAX_DZ = 0.024  # m
LOGGER = getLogger(__name__)
STATE_FORMAT = re.compile(
    r"wind:\s+([0-9.+-]+)\s+"
    r"tmp:\s+([0-9.+-]+)\s+"
    r"el:\s+([0-9.+-]+)\s+"
    r"time:\s+([0-9.+-]+)"
)
SUBREF_DX_FORMAT = re.compile(r"set\s+x\s+([0-9.+-]+)")
SUBREF_DZ_FORMAT = re.compile(r"set\s+z\s+([0-9.+-]+)")
TIME_FORMAT = "%y%m%d%H%M%S.%f"
TZ = ZoneInfo("Asia/Tokyo")


class Cosmos:
    """COSMOS client for the Nobeyama 45m telescope.

    Args:
        host: IP address of the COSMOS server.
        port: Port number of the COSMOS server.
        safe: Whether to raise an error before sending
            if the subreflector parameters are out of range.

    Example:
        ::

            from mao_45m.cosmos import Cosmos, Subref

            with Cosmos(host="127.0.0.1", port=11111) as cosmos:
                state = cosmos.receive_state()
                subref = cosmos.send_subref(dX=1.0, dZ=2.0)

    """

    def __init__(
        self,
        *,
        host: str = "127.0.0.1",
        port: int = 11111,
        safe: bool = True,
    ) -> None:
        self.sock = get_socket(host=host, port=port)
        self.safe = safe

    def send_subref(
        self,
        *,
        cmd: str = "pushsub",
        dX: float,
        dZ: float,
    ) -> "Subref":
        """Send the subreflector parameters of the Nobeyama 45m telescope.

        Args:
            cmd: Command name or path to send the parameters.
            dX: Offset (in m) from the X cylinder position
                optimized for the gravity deformation correction.
            dZ: Offset (in m) from the Z cylinder positions (Z1 = Z2)
                optimized for the gravity deformation correction.

        Returns:
            Current subreflector parameters received from COSMOS.

        """
        # check if dX and dZ are within the acceptable ranges
        if abs(dX) > ABSMAX_DX:
            if self.safe:
                raise ValueError(f"{dX=} is out of range (|dX| <= {ABSMAX_DX}).")
            else:
                LOGGER.warning(f"{dX=} is out of range (|dX| <= {ABSMAX_DX}).")

        if abs(dZ) > ABSMAX_DZ:
            if self.safe:
                raise ValueError(f"{dZ=} is out of range (|dZ| <= {ABSMAX_DZ}).")
            else:
                LOGGER.warning(f"{dZ=} is out of range (|dZ| <= {ABSMAX_DZ}).")

        # dX after m to mm conversion will be sent
        cmd_dX = f"{cmd} x {1e3 * dX}"
        LOGGER.debug(cmd_dX)
        self.sock.send((cmd_dX + "\n").encode())
        resp_dX = self.sock.recv(64)

        # dZ after m to mm conversion will be sent
        cmd_dZ = f"{cmd} z {1e3 * dZ}"
        LOGGER.debug(cmd_dZ)
        self.sock.send((cmd_dZ + "\n").encode())
        resp_dZ = self.sock.recv(64)

        # dX and dZ after mm to m will be stored
        return Subref.from_cosmos(resp_dX, resp_dZ)

    def receive_state(self, *, cmd: str = "pullwte") -> "State":
        """Receive the state of the Nobeyama 45m telescope.

        Args:
            cmd: Command name or path to receive the state.

        Returns:
            Current state of the Nobeyama 45m telescope.

        """

        LOGGER.debug(cmd)
        self.sock.send((cmd + "\n").encode())
        resp = self.sock.recv(64)

        return State.from_cosmos(resp)

    def __enter__(self) -> Self:
        """Enter the context manager."""
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        """Exit the context manager."""
        self.sock.close()


@dataclass(frozen=True)
class State:
    """Current state of the Nobeyama 45m telescope.

    Args:
        wind_speed: Wind speed (in m s^-1).
        temp: Ambient temperature (in deg C).
        elevation: Elevation angle of the telescope (in deg).
        time: Timestamp of the state (timezone-aware).

    """

    wind_speed: float
    temperature: float
    elevation: float
    time: datetime

    @classmethod
    def from_cosmos(cls, resp: bytes, /) -> Self:
        """Parse the current state from the COSMOS response."""
        if (match := STATE_FORMAT.search(resp.decode())) is None:
            raise ValueError("Could not parse the COSMOS response.")

        return cls(
            wind_speed=float(match[1]),
            temperature=float(match[2]),
            elevation=float(match[3]),
            time=datetime.strptime(match[4], TIME_FORMAT).replace(tzinfo=TZ),
        )


@dataclass(frozen=True)
class Subref:
    """Subreflector parameters of the Nobeyama 45m telescope.

    Args:
        dX: Offset (in m) from the X cylinder position
            optimized for the gravity deformation correction.
        dZ: Offset (in m) from the Z cylinder positions (Z1 = Z2)
            optimized for the gravity deformation correction.

    """

    dX: float
    dZ: float

    @classmethod
    def from_cosmos(cls, resp_dX: bytes, resp_dZ: bytes, /) -> Self:
        """Parse the subref parameters from the COSMOS response."""
        if (match_dX := SUBREF_DX_FORMAT.search(resp_dX.decode())) is None:
            raise ValueError("Could not parse the COSMOS response of dX.")

        if (match_dZ := SUBREF_DZ_FORMAT.search(resp_dZ.decode())) is None:
            raise ValueError("Could not parse the COSMOS response of dZ.")

        # dX and dZ after mm to m will be stored
        return cls(dX=1e-3 * float(match_dX[1]), dZ=1e-3 * float(match_dZ[1]))


def get_cosmos(
    *,
    host: str = "127.0.0.1",
    port: int = 11111,
    safe: bool = True,
) -> Cosmos:
    """Get a COSMOS client for the Nobeyama 45m telescope.

    Args:
        host: IP address of the COSMOS server.
        port: Port number of the COSMOS server.
        safe: Whether to raise an error before sending
            if the subreflector parameters are out of range.

    Returns:
        COSMOS client for the Nobeyama 45m telescope.

    """
    return Cosmos(host=host, port=port, safe=safe)


def receive(
    *,
    cmd: str = "pullwte",
    host: str = "127.0.0.1",
    port: int = 11111,
) -> State:
    """Receive the current state of the Nobeyama 45m telescope.

    Args:
        cmd: Command name or path to receive the state.
        host: IP address of the COSMOS server.
        port: Port number of the COSMOS server.

    Returns:
        Current state of the Nobeyama 45m telescope.

    """
    with Cosmos(host=host, port=port) as cosmos:
        return cosmos.receive_state(cmd=cmd)


def send(
    *,
    cmd: str = "pushsub",
    dX: float,
    dZ: float,
    host: str = "127.0.0.1",
    port: int = 11111,
    safe: bool = True,
) -> Subref:
    """Send the subreflector parameters of the Nobeyama 45m telescope.

    Args:
        cmd: Command name or path to send the parameters.
        dX: Offset (in m) from the X cylinder position
            optimized for the gravity deformation correction.
        dZ: Offset (in m) from the Z cylinder positions (Z1 = Z2)
            optimized for the gravity deformation correction.
        host: IP address of the COSMOS server.
        port: Port number of the COSMOS server.
        safe: Whether to raise an error before sending
            if the subreflector parameters are out of range.

    Returns:
        Current subreflector parameters received from COSMOS.

    """
    with Cosmos(host=host, port=port, safe=safe) as cosmos:
        return cosmos.send_subref(cmd=cmd, dX=dX, dZ=dZ)


def get_socket(*, host: str = "127.0.0.1", port: int = 11111) -> socket:
    """Get a socket object for TCP connection."""
    sock = socket(AF_INET, SOCK_STREAM)
    sock.connect((host, port))
    return sock



__all__ = [
    "Word",
    "get_channel_number",
    "get_corr_data",
    "get_elapsed_seconds",
    "get_frame_number",
    "get_integ_time",
    "get_ip_length",
    "get_reference_epoch",
    "get_samples",
    "get_sample_number",
    "get_time",
    "get_thread_id",
    "get_word",
]


# standard library
from dataclasses import dataclass


# dependencies
import numpy as np
import xarray as xr
from numpy.typing import NDArray
from typing_extensions import Self
from . import CHANS_PER_FRAME, CORR_DATA_BYTES, FRAMES_PER_SAMPLE, VDIF_HEAD_BYTES


@dataclass(frozen=True)
class Word:
    """Word parser for VDIF Head or Corr Head."""

    data: int

    def __getitem__(self, index: slice, /) -> int:
        """Get the slice of the word."""
        start, stop = index.start, index.stop
        return (self.data >> start) & ((1 << stop - start) - 1)

    @classmethod
    def from_bytes(cls, data: bytes, /) -> Self:
        """Create a Word parser from bytes."""
        return cls(int.from_bytes(data, "little"))


def get_channel_number(frame: bytes, /) -> int:
    """Get the channel number (1: CH0-255, ..., 64: CH16128-16383)."""
    return get_word(frame[:VDIF_HEAD_BYTES], 4)[16:24]


def get_corr_data(frame: bytes, /) -> NDArray[np.complex128]:
    """Get the complex auto or cross-correlation data."""
    data = np.frombuffer(frame[-CORR_DATA_BYTES:], dtype=np.int16)
    return data[0::2] + data[1::2] * 1j


def get_elapsed_seconds(frame: bytes, /) -> int:
    """Get the elapsed seconds from the reference epoch."""
    return get_word(frame[:VDIF_HEAD_BYTES], 0)[0:30]


def get_frame_number(frame: bytes, /) -> int:
    """Get the frame number within a second (0, 1, ...)."""
    return get_word(frame[:VDIF_HEAD_BYTES], 1)[0:24]


def get_integ_time(frame: bytes, /) -> float:
    """Get the integration time in seconds (0.005 or 0.010)."""
    return get_ip_length(frame) / 1000


def get_ip_length(frame: bytes, /) -> int:
    """Get the IP length (i.e. integration time) in ms (5 or 10)."""
    return get_word(frame[:VDIF_HEAD_BYTES], 4)[0:8]


def get_reference_epoch(frame: bytes, /) -> int:
    """Get the reference epoch (0: 2000 Jun, ..., 63: 2031 Jul.)."""
    return get_word(frame[:VDIF_HEAD_BYTES], 1)[24:30]


def get_samples(frames: list[bytes], /) -> xr.DataArray:
    """Get VDIF samples (time x chan) from VDIF frames."""
    data = []
    ids = []
    times = []

    for frame in frames:
        data.append(get_corr_data(frame))
        ids.append(get_channel_number(frame))
        times.append(get_time(frame))

    da = (
        xr.DataArray(
            data,
            dims=("time", "chan"),
            coords={
                "chan": np.arange(CHANS_PER_FRAME),
                "id": ("time", ids),
                "time": times,
            },
            attrs={
                "ip_length": get_ip_length(frames[0]),
            },
        )
        .set_index(temp=("id", "time"))
        .unstack("temp")
        .stack(temp=("chan", "id"), create_index=False)
    )

    return (
        da.assign_coords(temp=CHANS_PER_FRAME * (da.id - 1) + da.chan)
        .drop_vars(("chan", "id"))
        .rename(temp="chan")
        .sortby("time")
        .sortby("chan")
        .dropna("time")
    )


def get_sample_number(frame: bytes, /) -> int:
    """Get the sample number within a second (0, 1, ...)."""
    return get_frame_number(frame) // FRAMES_PER_SAMPLE


def get_time(frame: bytes, /) -> np.datetime64:
    """Get the recorded time of the frame in UTC."""
    return (
        np.datetime64("2000")
        + np.timedelta64(6 * get_reference_epoch(frame), "M")
        + np.timedelta64(get_elapsed_seconds(frame), "s")
        + np.timedelta64(get_sample_number(frame) * get_ip_length(frame), "ms")
        + np.timedelta64(0, "ns")
    )


def get_thread_id(frame: bytes, /) -> int:
    """Get the thread ID (1: IF 1x1, 2: IF 2x2, 5: IF 1x2, ...)."""
    return get_word(frame[:VDIF_HEAD_BYTES], 3)[16:26]


def get_word(head: bytes, n: int, /) -> Word:
    """Get the n-th word parser of a VDIF Head or Corr Head."""
    return Word.from_bytes(head[4 * n : 4 * (n + 1)])



__all__ = [
    "Converter",
    "get_converter",
    "get_homologous_epl",
    "get_integral_gain",
    "get_measurement_matrix",
    "get_proportional_gain",
]


# standard library
from dataclasses import dataclass
from functools import cached_property
from logging import getLogger
from os import PathLike


# dependencies
import numpy as np
import pandas as pd
import xarray as xr
from ..cosmos import ABSMAX_DX, ABSMAX_DZ
from ..utils import to_timedelta


# constants
LOGGER = getLogger(__name__)
SECOND = np.timedelta64(1, "s")


@dataclass
class Converter:
    """EPL-to-subref control converter for the Nobeyama 45m telescope..

    Args:
        G: Homologous EPL (G; feed x elevation; in m).
        K_I: Integral gain (K_I; feed).
        K_P: Proportional gain (K_I; feed).
        M: Measurement matrix (M; feed x drive).
        control_period: Control period (float in s or string with units).
        epl_interval_tolerance: Acceptable fraction of EPL time interval
            relative to the control period (0.1 means +/- 10% allowance).
        range_ddX: Absolute range for ddX (in m).
        range_ddZ: Absolute range for ddZ (in m).
        last: Last estimated subreflector parameters.

    """

    G: xr.DataArray
    M: xr.DataArray
    K_I: xr.DataArray
    K_P: xr.DataArray
    control_period: np.timedelta64 | str | float = "0.5 s"
    epl_interval_tolerance: float = 0.1
    range_ddX: tuple[float, float] = (0.00005, 0.000375)  # m
    range_ddZ: tuple[float, float] = (0.00005, 0.000300)  # m
    last: xr.DataArray | None = None

    @cached_property
    def inv_MTM_MT(self) -> xr.DataArray:
        """Pre-calculated (M^T M)^-1 M^T (drive x feed)."""
        M_ = self.M.rename(drive="drive_")
        return get_inv(M_ @ self.M) @ M_.T

    def __call__(self, epl: xr.DataArray, epl_cal: xr.DataArray, /) -> xr.DataArray:
        """Convert EPL to subreflector control (u; drive; in m).

        Args:
            epl: EPL to be converted (feed; in m)
                with the telescope state information at that time.
            epl_cal: EPL at calibration (feed; in m; must be zero)
                with the telescope state information at that time.

        Returns:
            Estimated subreflector control.

        """
        depl: xr.DataArray = (
            epl
            - epl_cal.data
            - self.G.interp(elevation=epl.elevation.data)
            + self.G.interp(elevation=epl_cal.elevation.data)
        )
        LOGGER.info(
            "Raw "
            + ", ".join(
                f"EPL({feed})={1e3 * epl.sel(feed=feed):+.3f}mm"
                for feed in epl.feed.data
            )
        )
        LOGGER.info(
            "Cor "
            + ", ".join(
                f"EPL({feed})={1e3 * depl.sel(feed=feed):+.3f}mm"
                for feed in depl.feed.data
            )
        )

        m = self.inv_MTM_MT @ depl
        tc = float(to_timedelta(self.control_period) / SECOND)

        if self.last is None:
            u: xr.DataArray = (
                (-self.K_I * tc * m - self.K_P * m)
                .assign_coords(m=m.assign_attrs(units="m"))
                .assign_attrs(units="m")
                .rename("u")
            )

            if abs(dX := float(u.sel(drive="X"))) > ABSMAX_DX:
                LOGGER.warning(f"{dX=} is out of range (|dX| <= {ABSMAX_DX}).")
                return self.on_failure(u)

            if abs(dZ := float(u.sel(drive="Z"))) > ABSMAX_DZ:
                LOGGER.warning(f"{dZ=} is out of range (|dZ| <= {ABSMAX_DZ}).")
                return self.on_failure(u)

            return self.on_success(u)

        u: xr.DataArray = (
            (self.last - self.K_I * tc * m - self.K_P * (m - self.last.m))
            .assign_coords(m=m.assign_attrs(units="m"))
            .assign_attrs(units="m")
            .rename("u")
        )
        dt = (u.time - self.last.time) / SECOND

        if abs(dX := float(u.sel(drive="X"))) > ABSMAX_DX:
            LOGGER.warning(f"{dX=} is out of range (|dX| <= {ABSMAX_DX}).")
            return self.on_failure(u)

        if abs(dZ := float(u.sel(drive="Z"))) > ABSMAX_DZ:
            LOGGER.warning(f"{dZ=} is out of range (|dZ| <= {ABSMAX_DZ}).")
            return self.on_failure(u)

        if abs(ddX := dX - float(self.last.sel(drive="X"))) < self.range_ddX[0]:
            LOGGER.warning(f"{ddX=} is out of range (|ddX| >= {self.range_ddX[0]}).")
            return self.on_failure(u)

        if abs(ddX) > self.range_ddX[1]:
            LOGGER.warning(f"{ddX=} is out of range (|ddX| <= {self.range_ddX[1]}).")
            return self.on_failure(u)

        if abs(ddZ := dZ - float(self.last.sel(drive="Z"))) < self.range_ddZ[0]:
            LOGGER.warning(f"{ddZ=} is out of range (|ddZ| >= {self.range_ddZ[0]}).")
            return self.on_failure(u)

        if abs(ddZ) > self.range_ddZ[1]:
            LOGGER.warning(f"{ddZ=} is out of range (|ddZ| <= {self.range_ddZ[1]}).")
            return self.on_failure(u)

        if dt < (dt_min := tc * (1 - self.epl_interval_tolerance)):
            LOGGER.warning(f"{dt=} is out of range (dt >= {dt_min}).")
            return self.on_failure(u)

        if dt > (dt_max := tc * (1 + self.epl_interval_tolerance)):
            LOGGER.warning(f"{dt=} is out of range (dt <= {dt_max}).")
            return self.on_failure(u)

        return self.on_success(u)

    def on_success(self, current: xr.DataArray, /) -> xr.DataArray:
        """Replace the last subreflector control with current one."""
        LOGGER.info(
            ", ".join(
                f"u({drive})={1e3 * current.sel(drive=drive):+.3f}mm"
                for drive in current.drive.data
            )
        )
        LOGGER.info(
            ", ".join(
                f"m({drive})={1e3 * current.m.sel(drive=drive):+.3f}mm"
                for drive in current.drive.data
            )
        )
        self.last = current
        return current

    def on_failure(self, current: xr.DataArray, /) -> xr.DataArray:
        """Replace the last subreflector control's time with current one."""
        if self.last is None:
            m_zero = xr.zeros_like(current.m)
            u_zero = xr.zeros_like(current)
            return self.on_success(u_zero.assign_coords(m=m_zero))
        else:
            return self.on_success(self.last.assign_coords(time=current.time))


def get_converter(
    *,
    control_period: np.timedelta64 | str | float = "0.5 s",
    epl_interval_tolerance: float = 0.1,
    feed_model: PathLike[str] | str,
    integral_gain_dX: float = 0.1,
    integral_gain_dZ: float = 0.1,
    proportional_gain_dX: float = 0.1,
    proportional_gain_dZ: float = 0.1,
    range_ddX: tuple[float, float] = (0.00005, 0.000375),  # m
    range_ddZ: tuple[float, float] = (0.00005, 0.000300),  # m
) -> Converter:
    """Get an EPL-to-subref control converter for the Nobeyama 45m telescope.

    Args:
        control_period: Control period (float in s or string with units).
        epl_interval_tolerance: Acceptable fraction of EPL time interval
            relative to the control period (0.1 means +/- 10% allowance).
        feed_model: Path to the feed model CSV file.
        integral_gain_dX: Integral gain for the estimated dX.
        integral_gain_dZ: Integral gain for the estimated dZ.
        proportional_gain_dX: Proportional gain for the estimated dX.
        proportional_gain_dZ: Proportional gain for the estimated dZ.
        range_ddX: Absolute range for ddX (in m).
        range_ddZ: Absolute range for ddZ (in m).

    Returns:
        EPL-to-subref control converter.

    """
    return Converter(
        G=get_homologous_epl(feed_model),
        M=get_measurement_matrix(feed_model),
        K_I=get_integral_gain(integral_gain_dX, integral_gain_dZ),
        K_P=get_proportional_gain(proportional_gain_dX, proportional_gain_dZ),
        control_period=control_period,
        epl_interval_tolerance=epl_interval_tolerance,
        range_ddX=range_ddX,
        range_ddZ=range_ddZ,
    )


def get_homologous_epl(
    feed_model: PathLike[str] | str,
    /,
    *,
    elevation_step: float = 0.01,
) -> xr.DataArray:
    """Get the homologous EPL (G; feed x elevation; in m) from given feed model.

    Args:
        feed_model: Path to the feed model CSV file.
        elevation_step: Elevation step size (in deg) for calculation.

    Returns:
        Homologous EPL (G; feed x elevation; in m).

    """
    df = pd.read_csv(feed_model, comment="#", index_col=0, skipinitialspace=True)

    a = xr.DataArray(
        df["homologous_EPL_A"],
        dims="feed",
        coords={"feed": df.index},
        attrs={"units": "m"},
    )
    b = xr.DataArray(
        df["homologous_EPL_B"],
        dims="feed",
        coords={"feed": df.index},
        attrs={"units": "deg"},
    )
    c = xr.DataArray(
        df["homologous_EPL_C"],
        dims="feed",
        coords={"feed": df.index},
        attrs={"units": "m"},
    )
    elevation = xr.DataArray(
        data := np.arange(0, 90.0 + elevation_step, elevation_step),
        dims="elevation",
        coords={"elevation": data},
        attrs={"units": "deg"},
    )

    with xr.set_options(keep_attrs=True):
        return (a * np.sin(np.deg2rad(elevation - b)) + c).rename("G")


def get_integral_gain(dX: float, dZ: float, /) -> xr.DataArray:
    """Get the integral gain (K_I; feed) from given values."""
    return xr.DataArray(
        data=[dX, dZ],
        dims="drive",
        coords={"drive": ["X", "Z"]},
        name="K_I",
    )


def get_inv(X: xr.DataArray, /) -> xr.DataArray:
    """Get the inverse of given two-dimensional DataArray."""
    return X.copy(data=np.linalg.inv(X.data.T)).T


def get_measurement_matrix(feed_model: PathLike[str] | str, /) -> xr.DataArray:
    """Get the measurement matrix (M; feed x drive) from given feed model.

    Args:
        feed_model: Path to the feed model CSV file.

    Returns:
        Measurement matrix (M; feed x drive).

    """
    df = pd.read_csv(feed_model, comment="#", index_col=0, skipinitialspace=True)

    return xr.DataArray(
        [df["EPL_over_dX"], df["EPL_over_dZ"]],
        dims=["drive", "feed"],
        coords={
            "drive": ["X", "Z"],
            "feed": df.index,
        },
        name="M",
    ).T


def get_proportional_gain(dX: float, dZ: float, /) -> xr.DataArray:
    """Get the proportional gain (K_P; feed) from given values."""
    return xr.DataArray(
        data=[dX, dZ],
        dims="drive",
        coords={"drive": ["X", "Z"]},
        name="K_P",
    )
